use super::PlayerAgent;
use crate::util::Symb;
use rand::Rng;
use std::num::NonZeroU8;

/// A simple "operator diffusion" MINIMIZER agent.
///
/// Tries to keep operators as far apart as possible, denying large numbers.
pub struct DiffuseAgent {}

impl DiffuseAgent {
	/// Place a symbol on the board.
	/// Assumes `symb` is not already on the board
	fn step_symb(&self, board: &mut crate::board::Board, symb: Symb) {
		if board.contains(symb) {
			panic!("Called `step_symb` with a symbol that's already on the board!")
		}

		// Fill distance array with largest possible value
		let mut dist = [board.size() + 1; 11];

		// Set up initial distances
		dist[0] = 1;
		*dist.last_mut().unwrap() = 1;
		for (i, o) in board.iter().enumerate() {
			if let Some((s, _)) = o {
				if s.is_op() {
					dist[i] = 0
				}
			}
		}

		let mut did_something = true;
		while did_something {
			did_something = false;
			for i in 1..(dist.len() - 1) {
				let l = dist[i - 1];
				let r = dist[i + 1];

				let new = (l + 1).min(r + 1);
				if new < dist[i] {
					did_something = true;
					dist[i] = new;
				}
			}
		}
		let mut max_dist = *dist.iter().max().unwrap();

		'outer: loop {
			for i in 0..11 {
				if dist[i] >= max_dist {
					if board.play(i, symb) {
						break 'outer;
					};
				}
			}
			max_dist -= 1;
		}
	}

	/// Called when we're out of symbols
	fn step_number(&self, board: &mut crate::board::Board) {
		let mut rng = rand::thread_rng();
		let n = board.size();

		let mut c = rng.gen_range(0..n);
		let mut s = {
			let n = rng.gen_range(0..=9);
			if n == 0 {
				Symb::Zero
			} else {
				Symb::Number(NonZeroU8::new(n).unwrap())
			}
		};

		while !board.play(c, s) {
			c = rng.gen_range(0..n);
			s = {
				let n = rng.gen_range(0..=9);
				if n == 0 {
					Symb::Zero
				} else {
					Symb::Number(NonZeroU8::new(n).unwrap())
				}
			};
		}
	}
}

impl PlayerAgent for DiffuseAgent {
	fn step(&mut self, board: &mut crate::board::Board) {
		let symb = [Symb::Minus, Symb::Times, Symb::Plus, Symb::Div]
			.iter()
			.filter(|x| !board.contains(**x))
			.next();

		// No symbols available, play a random number
		if symb.is_none() {
			self.step_number(board)
		} else {
			self.step_symb(board, *symb.unwrap())
		}
	}
}
