use anyhow::Result;
use std::{
	io::{stdin, stdout, StdoutLock, Write},
	num::NonZeroU8,
};
use termion::{
	color::{self},
	cursor::HideCursor,
	event::Key,
	input::TermRead,
	raw::IntoRawMode,
};

mod agents;
mod board;
mod util;
use board::Board;
use util::{Player, Symb};

fn play(
	stdout: &mut StdoutLock,
	player_max: bool,
	computer: &mut dyn agents::PlayerAgent,
) -> Result<Board> {
	let mut cursor = 0usize;
	let cursor_offset = 10usize - 1;
	let cursor_max = 10usize;

	let mut board = Board::new(if player_max {
		Player::Human
	} else {
		Player::Computer
	});

	let mut is_first = true;
	let mut print_board = true;
	'outer: loop {
		// Computer turn
		if board.current_player() == Player::Computer && !board.is_done() {
			computer.step(&mut board);
		}

		let min_color = if !player_max {
			Player::Human.color()
		} else {
			Player::Computer.color()
		};

		let max_color = if player_max {
			Player::Human.color()
		} else {
			Player::Computer.color()
		};

		if print_board {
			println!(
				"\r{}\r{}Min{}/{}Max{}  {}{}{}",
				" ".repeat(cursor_max + cursor_offset),
				color::Fg(min_color),
				color::Fg(color::Reset),
				color::Fg(max_color),
				color::Fg(color::Reset),
				if is_first { '╓' } else { '║' },
				board,
				if is_first { '╖' } else { '║' },
			);
			is_first = false;
		}

		print!(
			"\r{}╙{}{}{}{}{}╜",
			" ".repeat(cursor_offset),
			" ".repeat(cursor),
			color::Fg(board.current_player().color()),
			if board.is_done() { ' ' } else { '^' },
			color::Fg(color::Reset),
			" ".repeat(cursor_max - cursor),
		);
		stdout.flush()?;

		if board.is_done() {
			break;
		}

		// Player turn
		let stdin = stdin();
		for c in stdin.keys() {
			print_board = match c.unwrap() {
				Key::Right => {
					cursor = cursor_max.min(cursor + 1);
					false
				}
				Key::Left => {
					if cursor != 0 {
						cursor -= 1
					}
					false
				}
				Key::Char('q') => break 'outer,
				Key::Char('1') => board.play(cursor, Symb::Number(NonZeroU8::new(1).unwrap())),
				Key::Char('2') => board.play(cursor, Symb::Number(NonZeroU8::new(2).unwrap())),
				Key::Char('3') => board.play(cursor, Symb::Number(NonZeroU8::new(3).unwrap())),
				Key::Char('4') => board.play(cursor, Symb::Number(NonZeroU8::new(4).unwrap())),
				Key::Char('5') => board.play(cursor, Symb::Number(NonZeroU8::new(5).unwrap())),
				Key::Char('6') => board.play(cursor, Symb::Number(NonZeroU8::new(6).unwrap())),
				Key::Char('7') => board.play(cursor, Symb::Number(NonZeroU8::new(7).unwrap())),
				Key::Char('8') => board.play(cursor, Symb::Number(NonZeroU8::new(8).unwrap())),
				Key::Char('9') => board.play(cursor, Symb::Number(NonZeroU8::new(9).unwrap())),
				Key::Char('0') => board.play(cursor, Symb::Zero),
				Key::Char('+') => board.play(cursor, Symb::Plus),
				Key::Char('-') => board.play(cursor, Symb::Minus),
				Key::Char('*') => board.play(cursor, Symb::Times),
				Key::Char('/') => board.play(cursor, Symb::Div),
				_ => false,
			};
			continue 'outer;
		}
	}

	return Ok(board);
}

fn main() -> Result<()> {
	let stdout = HideCursor::from(stdout().into_raw_mode().unwrap());
	let mut stdout = stdout.lock();

	let mut agent = agents::DiffuseAgent {};
	let a = play(&mut stdout, true, &mut agent)?;
	if a.is_done() {
		println!(
			"\r\n{}Your score:{} {:.2}\n\n",
			color::Fg(Player::Human.color()),
			color::Fg(color::Reset),
			a.evaluate().unwrap()
		);
	} else {
		println!(
			"\r\n{}Quitting{}\r\n",
			color::Fg(color::Red),
			color::Fg(color::Reset),
		);
		return Ok(());
	}

	let mut agent = agents::RandomAgent {};
	let b = play(&mut stdout, false, &mut agent)?;
	if b.is_done() {
		println!(
			"\r\n{}Computer score:{} {:.2}\n\n",
			color::Fg(Player::Computer.color()),
			color::Fg(color::Reset),
			b.evaluate().unwrap()
		);
	} else {
		println!(
			"\r\n{}Quitting{}\r\n",
			color::Fg(color::Red),
			color::Fg(color::Reset),
		);
		return Ok(());
	}

	Ok(())
}
