use std::{fmt::Display, num::NonZeroU8};

use termion::color::{self, Color};

#[derive(PartialEq, Eq, Clone, Copy)]
pub enum Player {
	Human,
	Computer,
}

impl Player {
	pub fn invert(&mut self) {
		match self {
			Self::Human => *self = Self::Computer,
			Self::Computer => *self = Self::Human,
		}
	}

	pub fn color(&self) -> &dyn Color {
		match self {
			Player::Computer => &color::LightBlack,
			Player::Human => &color::Magenta,
		}
	}
}

impl Display for Player {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		let s = match self {
			Self::Computer => "Max",
			Self::Human => "Min",
		};
		write!(f, "{s}")
	}
}

#[derive(PartialEq, Eq, Clone, Copy)]
pub enum Symb {
	Number(NonZeroU8),
	Zero,
	Plus,
	Minus,
	Times,
	Div,
}

impl Display for Symb {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		match self {
			Self::Number(x) => x.fmt(f),
			Self::Zero => '0'.fmt(f),
			Self::Plus => '+'.fmt(f),
			Self::Minus => '-'.fmt(f),
			Self::Div => '÷'.fmt(f),
			Self::Times => '×'.fmt(f),
		}
	}
}

impl Symb {
	/// Is this symbol a plain binary operator?
	pub fn is_op(&self) -> bool {
		match self {
			Symb::Div | Symb::Plus | Symb::Times | Symb::Minus => true,
			_ => false,
		}
	}

	pub fn is_minus(&self) -> bool {
		self == &Self::Minus
	}

	pub fn from_char(c: &char) -> Option<Self> {
		match c {
			'1' => Some(Self::Number(NonZeroU8::new(1).unwrap())),
			'2' => Some(Self::Number(NonZeroU8::new(2).unwrap())),
			'3' => Some(Self::Number(NonZeroU8::new(3).unwrap())),
			'4' => Some(Self::Number(NonZeroU8::new(4).unwrap())),
			'5' => Some(Self::Number(NonZeroU8::new(5).unwrap())),
			'6' => Some(Self::Number(NonZeroU8::new(6).unwrap())),
			'7' => Some(Self::Number(NonZeroU8::new(7).unwrap())),
			'8' => Some(Self::Number(NonZeroU8::new(8).unwrap())),
			'9' => Some(Self::Number(NonZeroU8::new(9).unwrap())),
			'0' => Some(Self::Zero),
			'+' => Some(Self::Plus),
			'-' => Some(Self::Minus),
			'*' => Some(Self::Times),
			'/' => Some(Self::Div),
			'×' => Some(Self::Times),
			'÷' => Some(Self::Div),
			_ => None,
		}
	}
}
